/* copyright (c) 2007 magnus auvinen, see licence.txt for more info */
#include <base/system.h>
#include <base/math.hpp>
#include <base/vmath.hpp>

#include <math.h>
#include <engine/e_common_interface.h>
#include <game/mapitems.hpp>
#include <game/layers.hpp>
#include <game/collision.hpp>
#include <game/server/entities/character.hpp>

static TILE *tiles;
static int width = 0;
static int height = 0;

static int * teleports[NUM_TELEPORTS];
static int num_tele[NUM_TELEPORTS];

int col_width() { return width; }
int col_height() { return height; }

int col_init()
{
	width = layers_game_layer()->width;
	height = layers_game_layer()->height;
	tiles = (TILE *)map_get_data(layers_game_layer()->data);
	
	
	mem_zero(&num_tele, sizeof(num_tele));
	mem_zero(&teleports, sizeof(teleports));
	
	for(int i = 0; i < width*height; i++)
	{
		int tele_index = tiles[i].index - TELEPORT_OFFSET;
		if(tele_index >= 0 && tele_index < NUM_TELEPORTS)
		{
			if(tele_index&1) //from
				num_tele[tele_index]++;
		}
	}
	
	for(int i = 0; i < NUM_TELEPORTS; i++)
	{
		if(num_tele[i] > 0)
			teleports[i] = new int[num_tele[i]];
	}
	
	int cur_tele[NUM_TELEPORTS] = {0};
	for(int i = 0; i < width*height; i++)
	{
		int tele_index = tiles[i].index - TELEPORT_OFFSET;
		if(tele_index >= 0 && tele_index < NUM_TELEPORTS)
		{
			if(tele_index&1) //from
				teleports[tele_index][cur_tele[tele_index]++] = i;
		}
	}
	
	
	for(int i = 0; i < width*height; i++)
	{
		int index = tiles[i].index;
		
		if(index > 128)
			continue;
		
		if(index == TILE_DEATH)
			tiles[i].index = COLFLAG_DEATH;
		else if(index == TILE_SOLID)
			tiles[i].index = COLFLAG_SOLID;
		else if(index == TILE_NOHOOK)
			tiles[i].index = COLFLAG_SOLID|COLFLAG_NOHOOK;
		else if(index == TILE_HEALING)
			tiles[i].index = COLFLAG_HEALING;
		else if(index == TILE_POISON)
			tiles[i].index = COLFLAG_POISON;
		else
			tiles[i].index = 0;
		
		int tele_index = index - TELEPORT_OFFSET;
		if(tele_index >= 0 && tele_index < NUM_TELEPORTS)
			tiles[i].index = 128+tele_index;
	}
	
	return 1;
}


int col_get(int x, int y)
{
	int nx = clamp(x/32, 0, width-1);
	int ny = clamp(y/32, 0, height-1);
	
	if(tiles[ny*width+nx].index >= 128)
		return 0;
	return tiles[ny*width+nx].index;
}

int col_is_solid(int x, int y)
{
	return col_get(x, y)&COLFLAG_SOLID;
}


// TODO: rewrite this smarter!
int col_intersect_line(vec2 pos0, vec2 pos1, vec2 *out_collision, vec2 *out_before_collision, bool check_doors)
{
	float d = distance(pos0, pos1);
	vec2 last = pos0;
	
	for(float f = 0; f < d; f++)
	{
		float a = f/d;
		vec2 pos = mix(pos0, pos1, a);
		if(col_is_solid(round(pos.x), round(pos.y)) || (check_doors && col_is_door(round(pos.x), round(pos.y))))
		{
			if(out_collision)
				*out_collision = pos;
			if(out_before_collision)
				*out_before_collision = last;
			return col_get(round(pos.x), round(pos.y));
		}
		last = pos;
	}
	if(out_collision)
		*out_collision = pos1;
	if(out_before_collision)
		*out_before_collision = pos1;
	return 0;
}


vec2 col_teleport(int x, int y)
{
	int nx = clamp(x/32, 0, width-1);
	int ny = clamp(y/32, 0, height-1);
	
	int tele_index = tiles[ny*width+nx].index - 128;
	if(tele_index < 0 || tele_index >= NUM_TELEPORTS)
		return vec2(-1, -1);
	
	if(tele_index&1) //from
		return vec2(-1, -1);
	
	if(num_tele[tele_index+1] == 0)
		return vec2(-1, -1);
	
	vec2 closest = vec2(-1, -1);
	float closest_distance = 0.0f;
	for(int i = 0; i < num_tele[tele_index+1]; i++)
	{
		int dest = teleports[tele_index+1][i];
		vec2 dest_pos = vec2((dest%width)*32+16, (dest/width)*32+16);
		float d = distance(vec2(x, y), dest_pos);
		if(closest == vec2(-1, -1) || d < closest_distance)
		{
			closest_distance = d;
			closest = dest_pos;
		}
	}
	return closest;
}



void col_set_door(int start_x, int start_y, int end_x, int end_y)
{
	int nstart_x = clamp(start_x/32, 0, width-1);
	int nstart_y = clamp(start_y/32, 0, height-1);
	int nend_x = clamp(end_x/32, 0, width-1);
	int nend_y = clamp(end_y/32, 0, height-1);
	
	for(int nx = nstart_x; nx <= nend_x; nx++)
	{
		for(int ny = nstart_y; ny <= nend_y; ny++)
		{
			tiles[ny*width+nx].index |= COLFLAG_NOHOOK|COLFLAG_DOOR;
		}
	}
}

int col_is_door(int x, int y)
{
	return col_get(x, y)&COLFLAG_DOOR;
}
