#include <string.h> //strncmp

#include <engine/e_server_interface.h> //server_tick, server_tickspeed
#include <engine/e_config.h>

#include <game/mapitems.hpp>
#include <game/server/gamecontext.hpp>
#include <game/server/entities/flag.hpp>
#include <game/server/entities/pickup.hpp>
#include "exp.hpp"



GAMECONTROLLER_EXP::GAMECONTROLLER_EXP(VOTEOPTION * voteoption_first)
{
	gametype = "EXP";
	game_flags = GAMEFLAG_TEAMS | GAMEFLAG_FLAGS;
	
	cur_flag = 0;
	cur_turret = 0;
	cur_mine = 0;
	cur_trap = 0;
	cur_door = 0;
	
	for(int i = 0; i < BOT_LEVELS; i++)
		num_bot_spawns[i] = 0;
	
	for(int i = 0; i < MAX_CHECKPOINTS; i++)
		flags_cp[i] = NULL;
	flag_end = NULL;
	
	boss.exist = false;
	for(int i = 0; i < 3; i++)
		boss.shield_icons[i] = NULL;
	
	first_option = voteoption_first;
	
	//force config
	config.sv_max_clients = 6;
	config.sv_scorelimit = 1;
	config.sv_teamdamage = 0;
	
	fs_makedir("database");
}



bool GAMECONTROLLER_EXP::on_entity(int index, vec2 pos)
{
	int lvl = 0;
	if(index == ENTITY_SPAWN_BOT_LEVEL_1)
		lvl = 1;
	else if(index == ENTITY_SPAWN_BOT_LEVEL_2)
		lvl = 2;
	else if(index == ENTITY_SPAWN_BOT_LEVEL_3)
		lvl = 3;
	if(lvl != 0)
	{
		dbg_msg("exp", "Bot Spawn Level %d added (%d)", lvl, num_bot_spawns[lvl-1]);
		bot_spawns[lvl-1][num_bot_spawns[lvl-1]].pos = pos;
		bot_spawns[lvl-1][num_bot_spawns[lvl-1]].level = lvl;
		bot_spawns[lvl-1][num_bot_spawns[lvl-1]].spawned = false;
		bot_spawns[lvl-1][num_bot_spawns[lvl-1]].respawn_timer = (float)server_tick() - tuning.respawn_timer*server_tickspeed();
		num_bot_spawns[lvl-1]++;
	}
	
	if(index == ENTITY_SPAWN_BOSS)
	{
		if(boss.exist)
			dbg_msg("exp", "There can't be 2 boss");
		else
		{
			dbg_msg("exp", "Boss added");
			boss.exist = true;
			boss.spawn.pos = pos;
			boss.spawn.level = 4;
			boss.spawn.spawned = false;
			boss.spawn.respawn_timer = (float)server_tick() - tuning.respawn_timer*server_tickspeed();
		}
	}
	
	else if(index == ENTITY_FLAGSTAND_RED)
	{
		if(cur_flag < MAX_CHECKPOINTS)
		{
			dbg_msg("exp", "Checkpoint added (%d)", cur_flag);
			FLAG * f = new FLAG(0);
			f->stand_pos = pos;
			f->pos = pos;
			flags_cp[cur_flag++] = f;
			config.sv_scorelimit++;
		}
		return true;
	}
	else if(index == ENTITY_FLAGSTAND_BLUE)
	{
		dbg_msg("exp", "Blue Flag added");
		FLAG * f = new FLAG(1);
		f->stand_pos = pos;
		f->pos = pos;
		flag_end = f;
		return true;
	}
	
	else if(index == ENTITY_TURRET_LASER)
	{
		if(cur_turret < MAX_TURRETS)
		{
			dbg_msg("exp", "Laser Turret added (%d)", cur_turret);
			turrets[cur_turret].used = true;
			turrets[cur_turret].pos = pos;
			turrets[cur_turret].type = TURRET_TYPE_LASER;
			build_turret(cur_turret++);
		}
		else
			dbg_msg("exp", "Can't create Laser Turret : too many Turrets.");
		return true;
	}
	
	else if(index == ENTITY_TURRET_GUN)
	{
		if(cur_turret < MAX_TURRETS)
		{
			dbg_msg("exp", "Gun Turret added (%d)", cur_turret);
			turrets[cur_turret].used = true;
			turrets[cur_turret].pos = pos;
			turrets[cur_turret].type = TURRET_TYPE_GUN;
			build_turret(cur_turret++);
		}
		else
			dbg_msg("exp", "Can't create Gun Turret : too many Turrets.");
		return true;
	}
	
	else if(index == ENTITY_MINE)
	{
		if(cur_mine < MAX_MINES)
		{
			dbg_msg("exp", "Mine added (%d)", cur_mine);
			mines[cur_mine].used = true;
			mines[cur_mine].pos = vec2(pos.x, pos.y+14);
			build_mine(cur_mine++);
		}
		else
			dbg_msg("exp", "Can't create Mine : too many Mines.");
		return true;
	}
	
	else if(index == ENTITY_TRAP)
	{
		if(cur_trap < MAX_TRAPS)
		{
			dbg_msg("exp", "Trap added (%d)", cur_trap);
			traps[cur_trap].used = true;
			traps[cur_trap].pos = vec2(pos.x, pos.y-14);
			build_trap(cur_trap++);
		}
		else
			dbg_msg("exp", "Can't create Trap : too many Traps.");
		return true;
	}
	
	else if(index == ENTITY_DOOR_VERTICAL)
	{
		if(cur_door < MAX_DOORS)
		{
			dbg_msg("exp", "Vertical Door added (%d)", cur_door);
			doors[cur_door].used = true;
			doors[cur_door].pos = vec2(pos.x, pos.y-16);
			doors[cur_door].type = DOOR_TYPE_VERTICAL;
			build_door(cur_door++);
		}
		else
			dbg_msg("exp", "Can't create Vertical Door : too many Doors.");
		return true;
	}
	
	else if(index == ENTITY_DOOR_HORIZONTAL)
	{
		if(cur_door < MAX_DOORS)
		{
			dbg_msg("exp", "Horizontal Door added (%d)", cur_door);
			doors[cur_door].used = true;
			doors[cur_door].pos = vec2(pos.x-16, pos.y);
			doors[cur_door].type = DOOR_TYPE_HORIZONTAL;
			build_door(cur_door++);
		}
		else
			dbg_msg("exp", "Can't create Horizontal Door : too many Doors.");
		return true;
	}
	
	else if(GAMECONTROLLER::on_entity(index, pos))
		return true;
	
	return false;
}



void GAMECONTROLLER_EXP::tick()
{
	tick_save();
	
	tick_bots();
	
	tick_weapons();
	
	GAMECONTROLLER::tick();
}



void GAMECONTROLLER_EXP::on_chat(int client_id, int team, const char * msg)
{
	if(!strncmp(msg, "/info", 5) || !strncmp(msg, ".info", 5) || !strncmp(msg, "/help", 5))
	{
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
		game.send_chat_target(client_id, "                                        EXPlorer");
		game.send_chat_target(client_id, "");
		game.send_chat_target(client_id, "Version 1.1 | By Choupom");
		game.send_chat_target(client_id, "You have to explore the map, fight monsters, collect items...");
		game.send_chat_target(client_id, "Kill a monster to earn an Item (say /items for more info).");
		game.send_chat_target(client_id, "");
		game.send_chat_target(client_id, "Say /command for the command list.");
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
	}
	else if(!strncmp(msg, "/command", 8) || !strncmp(msg, "/cmd", 4))
	{
		game.send_chat_target(client_id, "");
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
		game.send_chat_target(client_id, "                                    COMMAND LIST");
		game.send_chat_target(client_id, "");
		game.send_chat_target(client_id, "/info : Get info about the MOD.");
		game.send_chat_target(client_id, "/items : Get info about the Items.");
		game.send_chat_target(client_id, "/new : Restart the game.");
		game.send_chat_target(client_id, "/bind : Learn how to bind a key to use an item.");
		game.send_chat_target(client_id, "/game : Show your Weapons, Kills, Armor, etc.");
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
	}
	else if(!strncmp(msg, "/items", 6))
	{
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
		game.send_chat_target(client_id, "                                           ITEMS");
		game.send_chat_target(client_id, "");
		game.send_chat_target(client_id, "Go in votes to use an item.");
		game.send_chat_target(client_id, "Weapons : You keep it when you have it.");
		game.send_chat_target(client_id, "Life : You can use it to respawn where you died.");
		game.send_chat_target(client_id, "Minor Potion : Use it to get full health.");
		game.send_chat_target(client_id, "Greater Potion : Use it to get full health and full armor.");
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
	}
	else if(!strncmp(msg, "/game", 5))
	{
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
		game.send_chat_target(client_id, "                                            GAME");
		game.send_chat_target(client_id, "");
		
		char buf[256] = {0};
		if(game.players[client_id]->game_exp.weapons == 0)
			str_format(buf, sizeof(buf), "Weapons : None", buf);
		for(int i = 1; i < NUM_WEAPONS; i++)
		{
			if(game.players[client_id]->game_exp.weapons & i)
			{
				if(buf[0] == 0) str_format(buf, sizeof(buf), "Weapons : %s", get_weapon_name(i));
				else str_format(buf, sizeof(buf), "%s, %s", buf, get_weapon_name(i));
			}
		}
		game.send_chat_target(client_id, buf);
		
		str_format(buf, sizeof(buf), "Armor : %d", game.players[client_id]->game_exp.armor_max);
		game.send_chat_target(client_id, buf);
		
		str_format(buf, sizeof(buf), "Kills : %d", game.players[client_id]->score);
		game.send_chat_target(client_id, buf);
		
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
	}
	
	else if(!strncmp(msg, "/bind", 5))
	{
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
		game.send_chat_target(client_id, "                              BIND DOCUMENTATION");
		game.send_chat_target(client_id, "");
		game.send_chat_target(client_id, "1) Open the Local Console (F1).");
		game.send_chat_target(client_id, "2) Type \"bind <key> say <item>\"");
		game.send_chat_target(client_id, "Replace <key> by the key you want to press.");
		game.send_chat_target(client_id, "Replace <item> by the item : life, minor or greater.");
		game.send_chat_target(client_id, "Example : \"bind l say life\"");
		game.send_chat_target(client_id, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
	}
	
	else if(!strncmp(msg, "/new", 4))
	{
		restart_client(client_id);
	}
	
	else if(!strncmp(msg, "/life", 5) || !strncmp(msg, "life", 4))
	{
		use(client_id, "Life");
	}

	else if(!strncmp(msg, "/minor", 6) || !strncmp(msg, "minor", 5))
	{
		use(client_id, "Minor Potion");
	}

	else if(!strncmp(msg, "/greater", 8) || !strncmp(msg, "greater", 7))
	{
		use(client_id, "Greater Potion");
	}
	
	else if(!strncmp(msg, "/", 1))
	{
		char buf[256];
		str_format(buf, sizeof(buf), "Unknown command : %s", msg);
		game.send_chat_target(client_id, buf);
	}
	else
		game.send_chat(client_id, team, msg);
}



void GAMECONTROLLER_EXP::on_connect(int client_id)
{
	game.send_chat_target(client_id, "Welcome in EXPlorer mod. Say /info for more info about EXPlorer.");
	load_client(client_id);
}



int GAMECONTROLLER_EXP::on_character_death(class CHARACTER * victim, class PLAYER * killer, int weapon)
{
	if(!victim->player->is_bot() && weapon != WEAPON_GAME)
	{
		if(victim->player->game_exp.items.lives > 0)
		{
			char buf[256];
			str_format(buf, sizeof(buf), "You have %d lives. Use a life to respawn where you died.", victim->player->game_exp.items.lives);
			game.send_broadcast(buf, victim->player->client_id);
		}
	}
	
	return GAMECONTROLLER::on_character_death(victim, killer, weapon);
}



const char * GAMECONTROLLER_EXP::get_weapon_name(int wid)
{
	if(wid == WEAPON_HAMMER)
		return "Hammer";
	else if(wid == WEAPON_GUN)
		return "Gun";
	else if(wid == WEAPON_SHOTGUN)
		return "Shotgun";
	else if(wid == WEAPON_GRENADE)
		return "Grenade";
	else if(wid == WEAPON_RIFLE)
		return "Rifle";
	else if(wid == WEAPON_NINJA)
		return "Ninja";
	else if(wid == WEAPON_KAMIKAZE)
		return "Kamikaze";
	else if(wid == WEAPON_FREEZER)
		return "Freezer";
	return "?";
}



void register_exp_commands()
{
	MACRO_REGISTER_COMMAND("weapons", "i", CFGFLAG_SERVER, con_get_weapons, 0, "");
	MACRO_REGISTER_COMMAND("teleflag", "ii", CFGFLAG_SERVER, con_teleflag, 0, "");
	MACRO_REGISTER_COMMAND("teleport", "ii", CFGFLAG_SERVER, con_teleport, 0, "");
}


void con_get_weapons(void *result, void *user_data)
{
	if(!game.controller->isEXP())
		return;
	
	int id = console_arg_int(result, 0);
	
	if(id < 0 || id >= MAX_CLIENTS || !game.players[id])
		return;
	if(game.players[id]->is_bot())
		return;
	
	for(int i = 0; i < NUM_WEAPONS; i++)
	{
		if(i == WEAPON_NINJA)
			continue;
		game.players[id]->get_weapon(i);
	}
}


void con_teleflag(void *result, void *user_data)
{
	if(!game.controller->isEXP())
		return;
	
	int id = console_arg_int(result, 0);
	int flag = console_arg_int(result, 1);
	
	if(id < 0 || id >= MAX_CLIENTS || !game.players[id])
		return;
	if(game.players[id]->is_bot())
		return;
	
	if(flag < 0 || flag > ((GAMECONTROLLER_EXP *)game.controller)->cur_flag)
		return;
	
	game.players[id]->game_exp.last_flag = flag;
	if(flag == 0)
		((GAMECONTROLLER_EXP *)game.controller)->restart_client(id);
	else if(game.players[id]->get_character())
		game.players[id]->get_character()->teleport(((GAMECONTROLLER_EXP *)game.controller)->flags_cp[flag-1]->pos);
}


void con_teleport(void *result, void *user_data)
{
	if(!game.controller->isEXP())
		return;
	
	int id1 = console_arg_int(result, 0);
	int id2 = console_arg_int(result, 1);
	
	if(id1 < 0 || id1 >= MAX_CLIENTS || !game.players[id1])
		return;
	if(game.players[id1]->is_bot() || !game.players[id1]->get_character())
		return;
	
	if(id2 < 0 || id2 >= MAX_CLIENTS || !game.players[id2])
		return;
	if(!game.players[id2]->get_character())
		return;
	
	game.players[id1]->get_character()->teleport(game.players[id2]->get_character()->pos);
}
