#include <game/server/gamecontroller.hpp>
#include <game/server/entities/flag.hpp>

#include "bots.hpp"
#include "save.hpp"
#include "environment.hpp"



struct VOTEOPTION
{
	VOTEOPTION *next;
	VOTEOPTION *prev;
	char command[1];
};

enum
{
	FLAG_VISIBLE_RADIUS=850
};



class GAMECONTROLLER_EXP : public GAMECONTROLLER
{
public:
	GAMECONTROLLER_EXP(VOTEOPTION * voteoption_first);
	virtual void tick();
	virtual int on_character_death(class CHARACTER * victim, class PLAYER * killer, int weapon);
	virtual bool isEXP() const	{	return true;	}
	
	void on_chat(int client_id, int team, const char * msg);
	void on_connect(int client_id);
	
	void tick_bots(); void tick_save(); void tick_weapons();
	
	//entities :
	virtual bool on_entity(int index, vec2 pos);
	int cur_flag;
	int cur_turret;
	int cur_mine;
	int cur_trap;
	int cur_door;
	
	//bots variables :
	BOT_SPAWN bot_spawns[BOT_LEVELS][MAX_BOT_SPAWNS];
	int num_bot_spawns[BOT_LEVELS];
	BOSS boss;
	
	//bots functions :
	int bot_can_spawn();
	void bot_spawn(BOT_SPAWN * spawn);
	void remove_bot(int id, bool killed);
	
	//save functions :
	void save_client(int id);
	void load_client(int id);
	void restart_client(int id);
	void start_client(int id);
	void stop_client(int id);
	void update_game(int id);
	void get_filename(int id, char buffer[256]);
	
	//flags :
	FLAG * flags_cp[MAX_CHECKPOINTS];
	FLAG * flag_end;
	
	//environment variables :
	TURRET turrets[MAX_TURRETS];
	MINE mines[MAX_MINES];
	TRAP traps[MAX_TRAPS];
	DOOR doors[MAX_DOORS];
	
	//environment functions :
	void build_turret(int t);
	void hit_turret(int t, int owner, int dmg);
	void destroy_turret(int t, int killer);
	void freeze_turret(int t);
	void build_mine(int m);
	void destroy_mine(int m);
	void build_trap(int t);
	void build_door(int d);
	
	//items functions :
	void send_items(int client_id);
	void send_item(int client_id, int item);
	bool use(int client_id, const char * command);

	//other/all :
	VOTEOPTION * first_option;
	const char * get_weapon_name(int wid);
};



void register_exp_commands();
void con_get_weapons(void *result, void *user_data);
void con_teleflag(void *result, void *user_data);
void con_teleport(void *result, void *user_data);
