#include <new>

#include <engine/e_server_interface.h>

#include "player.hpp"
#include "gamecontext.hpp"
#include "gamemodes/exp/exp.hpp"
#include "gamemodes/exp/treasures.hpp"
#include "entities/flag.hpp"



MACRO_ALLOC_POOL_ID_IMPL(PLAYER, MAX_CLIENTS)

PLAYER::PLAYER(int client_id)
{
	respawn_tick = server_tick();
	character = 0;
	this->client_id = client_id;
}

PLAYER::~PLAYER()
{
	delete character;
	character = 0;
}

void PLAYER::tick()
{
	if(game.controller->isEXP())
		server_setclientscore(client_id, game_exp.last_flag);
	else
		server_setclientscore(client_id, score);
	
	// do latency stuff
	if(!(game.controller->isEXP() && is_bot()))
	{
		CLIENT_INFO info;
		if(server_getclientinfo(client_id, &info))
		{
			latency.accum += info.latency;
			latency.accum_max = max(latency.accum_max, info.latency);
			latency.accum_min = min(latency.accum_min, info.latency);
		}

		if(server_tick()%server_tickspeed() == 0)
		{
			latency.avg = latency.accum/server_tickspeed();
			latency.max = latency.accum_max;
			latency.min = latency.accum_min;
			latency.accum = 0;
			latency.accum_min = 1000;
			latency.accum_max = 0;
		}
	}
	
	int sec = (game.controller->isEXP() ? 10 : 3);
	if(!character && die_tick+server_tickspeed()*sec <= server_tick())
		spawning = true;

	if(character)
	{
		if(character->alive)
		{
			view_pos = character->pos;
		}
		else
		{
			delete character;
			character = 0;
		}
	}
	else if(spawning && respawn_tick <= server_tick())
		try_respawn();
}

void PLAYER::snap(int snapping_client)
{
	/*if(is_bot())
	{
		if(!get_character())
			return;
		if(character->networkclipped(snapping_client))
			return;
	}*/
	
	NETOBJ_CLIENT_INFO *client_info = (NETOBJ_CLIENT_INFO *)snap_new_item(NETOBJTYPE_CLIENT_INFO, client_id, sizeof(NETOBJ_CLIENT_INFO));
	if(game.controller->isEXP() && is_bot())
		str_to_ints(&client_info->name0, 6, get_monster_name());
	else
		str_to_ints(&client_info->name0, 6, server_clientname(client_id));
	str_to_ints(&client_info->skin0, 6, skin_name);
	client_info->use_custom_color = use_custom_color;
	client_info->color_body = color_body;
	client_info->color_feet = color_feet;

	NETOBJ_PLAYER_INFO *info = (NETOBJ_PLAYER_INFO *)snap_new_item(NETOBJTYPE_PLAYER_INFO, client_id, sizeof(NETOBJ_PLAYER_INFO));
	
	if(game.controller->isEXP() && is_bot())
	{
		info->latency = 0;
		info->latency_flux = 0;
	}
	else
	{
		info->latency = latency.min;
		info->latency_flux = latency.max-latency.min;
	}
	info->local = 0;
	info->cid = client_id;
	
	if(game.controller->isEXP())
	{
		if(is_bot())
			info->score = 0;
		else
			info->score = game_exp.last_flag;
	}
	else
		info->score = score;
	
	info->team = team;

	if(client_id == snapping_client)
		info->local = 1;	
}

void PLAYER::on_disconnect()
{
	kill_character(WEAPON_GAME);
	
	//game.controller->on_player_death(&game.players[client_id], 0, -1);
		
	char buf[512];
	str_format(buf, sizeof(buf),  "%s has left the game", server_clientname(client_id));
	game.send_chat(-1, GAMECONTEXT::CHAT_ALL, buf);

	dbg_msg("game", "leave player='%d:%s'", client_id, server_clientname(client_id));
}

void PLAYER::on_predicted_input(NETOBJ_PLAYER_INPUT *new_input)
{
	CHARACTER *chr = get_character();
	if(chr)
		chr->on_predicted_input(new_input);
}

void PLAYER::on_direct_input(NETOBJ_PLAYER_INPUT *new_input)
{
	CHARACTER *chr = get_character();
	if(chr)
		chr->on_direct_input(new_input);

	if(!chr && team >= 0 && (new_input->fire&1))
		spawning = true;
	
	if(!chr && team == -1)
		view_pos = vec2(new_input->target_x, new_input->target_y);
}

CHARACTER *PLAYER::get_character()
{
	if(character && character->alive)
		return character;
	return 0;
}

void PLAYER::kill_character(int weapon)
{
	if(character)
	{
		character->die(client_id, weapon);
		delete character;
		character = 0;
	}
}

void PLAYER::respawn()
{
	if(team > -1)
		spawning = true;
}

void PLAYER::set_team(int new_team)
{
	// clamp the team
	new_team = game.controller->clampteam(new_team);
	if(team == new_team)
		return;
	
	if(game.controller->isEXP())
	{
		if(new_team == -1)
			((GAMECONTROLLER_EXP *)game.controller)->update_game(client_id);
		else if(new_team == 0)
			game_exp.enter_tick = server_tick();
	}
	
	char buf[512];
	str_format(buf, sizeof(buf), "%s joined the %s", server_clientname(client_id), game.controller->get_team_name(new_team));
	game.send_chat(-1, GAMECONTEXT::CHAT_ALL, buf); 
	
	kill_character(WEAPON_GAME);
	team = new_team;
	score = 0;
	dbg_msg("game", "team_join player='%d:%s' team=%d", client_id, server_clientname(client_id), team);
	
	game.controller->on_player_info_change(game.players[client_id]);
}

void PLAYER::try_respawn()
{
	if(game.controller->isEXP())
	{
		if(game_exp.last_flag == 0)
		{
			vec2 spawnpos = vec2(0.0f, 0.0f);
			if(!game.controller->can_spawn(this, &spawnpos))
				return;
			
			game_exp.enter_tick = server_tick();
			load_game(spawnpos, game_exp.last_flag, score, 0, game_exp.armor_max, game_exp.weapons, game_exp.items, game_exp.boss_hitter, game_exp.boss_killer);
		}
		else if(team != -1)
		{
			load_game(((GAMECONTROLLER_EXP *)game.controller)->flags_cp[game_exp.last_flag-1]->pos, game_exp.last_flag,
				score, game_exp.time, game_exp.armor_max, game_exp.weapons, game_exp.items, game_exp.boss_hitter, game_exp.boss_killer);
		}
		game.send_broadcast("", client_id);
	}
	else
	{
		vec2 spawnpos = vec2(100.0f, -60.0f);
		if(!game.controller->can_spawn(this, &spawnpos))
			return;
		// check if the position is occupado
		ENTITY *ents[2] = {0};
		int num_ents = game.world.find_entities(spawnpos, 64, ents, 2, NETOBJTYPE_CHARACTER);
		
		if(num_ents == 0)
		{
			spawning = false;
			character = new(client_id) CHARACTER();
			character->spawn(this, spawnpos, team);
			game.create_playerspawn(spawnpos);
		}
	}
}





void PLAYER::make_bot(BOT_SPAWN * spawn)
{
	character = new(client_id) CHARACTER();
	team = 1;
	character->spawn(game.players[client_id], spawn->pos, 1);
	game.create_playerspawn(spawn->pos);
	bot_level = spawn->level;
	bot_spawn = spawn;
	
	character->health = max_health();
	character->armor = max_armor();
	
	if(bot_level == 1)
	{
		str_copy(skin_name, "coala", sizeof(skin_name));
	}
	else if(bot_level == 2)
	{
		str_copy(skin_name, "kamikaze", sizeof(skin_name));
		character->weapons[WEAPON_HAMMER].got = false;
		character->weapons[WEAPON_GUN].got = false;
		character->weapons[WEAPON_KAMIKAZE].got = true;
		character->weapons[WEAPON_KAMIKAZE].ammo = -1;
		character->queued_weapon = WEAPON_KAMIKAZE;
	}
	else if(bot_level == 3)
	{
		str_copy(skin_name, "guard", sizeof(skin_name));
		character->weapons[WEAPON_SHOTGUN].got = true;
		character->weapons[WEAPON_SHOTGUN].ammo = 10;
	}
	else if(bot_level == 4)
	{
		str_copy(skin_name, "boss_shield", sizeof(skin_name));
		character->weapons[WEAPON_GUN].got = false;
		character->weapons[WEAPON_FREEZER].got = true;
		character->weapons[WEAPON_FREEZER].ammo = -1;
	}
}


int PLAYER::max_health()
{
	if(bot_level == 1)
		return 10;
	else if(bot_level == 2)
		return 8;
	else if(bot_level == 3)
		return 10;
	else if(bot_level == 4)
		return 200;
	return 10;
}
int PLAYER::max_armor()
{
	if(bot_level == 1)
		return 0;
	else if(bot_level == 2)
		return 0;
	else if(bot_level == 3)
		return 5;
	else if(bot_level == 4)
		return 0;
	return 10;
}


const char * PLAYER::get_monster_name()
{
	int life = character->health + character->armor;
	int max_life = max_health() + max_armor();
	float coef = (float)life / (float)max_life;
	
	if(coef > 0.8f)
		return "|-----|";
	else if(coef > 0.6f)
		return "|----|";
	else if(coef > 0.4f)
		return "|---|";
	else if(coef > 0.2f)
		return "|--|";
	else
		return "|-|";
}



void PLAYER::load_game(vec2 spawnpos, int flag, int kills, int time, int armor, flags w, ITEMS items, bool b_hitter, bool b_killer)
{
	spawning = false;
	character = new(client_id) CHARACTER();
	character->spawn(game.players[client_id], spawnpos, 1);
	game.create_playerspawn(spawnpos);
	game_exp.armor_max = armor;
	character->armor = armor;
	game_exp.last_flag = flag;
	game_exp.kills = kills;
	score = kills;
	game_exp.time = time;
	if(w & WEAPON_GUN)
	{
		character->weapons[WEAPON_GUN].got = true;
		character->weapons[WEAPON_GUN].ammo = data->weapons.id[WEAPON_GUN].maxammo;
	}
	if(w & WEAPON_SHOTGUN)
	{
		character->weapons[WEAPON_SHOTGUN].got = true;
		character->weapons[WEAPON_SHOTGUN].ammo = data->weapons.id[WEAPON_SHOTGUN].maxammo;
	}
	if(w & WEAPON_GRENADE)
	{
		character->weapons[WEAPON_GRENADE].got = true;
		character->weapons[WEAPON_GRENADE].ammo = data->weapons.id[WEAPON_GRENADE].maxammo;
	}
	if(w & WEAPON_RIFLE)
	{
		character->weapons[WEAPON_RIFLE].got = true;
		character->weapons[WEAPON_RIFLE].ammo = data->weapons.id[WEAPON_RIFLE].maxammo;
	}
	if(w & WEAPON_KAMIKAZE)
	{
		character->weapons[WEAPON_KAMIKAZE].got = true;
		character->weapons[WEAPON_KAMIKAZE].ammo = -1;
	}
	if(w & WEAPON_FREEZER)
	{
		character->weapons[WEAPON_FREEZER].got = true;
		character->weapons[WEAPON_FREEZER].ammo = -1;
	}
	game_exp.weapons = w;
	game_exp.items = items;
	game_exp.boss_hitter = b_hitter;
	game_exp.boss_killer = b_killer;
}



void PLAYER::get_weapon(int wid)
{
	if(game_exp.weapons & wid)
		return;
	
	if(wid == WEAPON_RIFLE && game_exp.weapons & WEAPON_FREEZER)
		return;
	if(wid == WEAPON_FREEZER)
	{
		game_exp.weapons -= WEAPON_RIFLE;
		if(character)
			character->weapons[WEAPON_RIFLE].got = false;
	}
	
	game_exp.weapons += wid;
	
	if(character)
	{
		character->weapons[wid].got = true;
		
		if(wid == WEAPON_HAMMER || wid == WEAPON_KAMIKAZE || wid == WEAPON_FREEZER)
			character->weapons[wid].ammo = -1;
		else
			character->weapons[wid].ammo = data->weapons.id[wid].maxammo;
	}
}
