#ifndef GAME_SERVER_PLAYER_H
#define GAME_SERVER_PLAYER_H


#include <game/flags.hpp>
#include "entities/character.hpp"
#include "gamemodes/exp/treasures.hpp"


class PLAYER
{
	friend class GAMECONTROLLER_EXP;
	
	MACRO_ALLOC_POOL_ID()
private:
	CHARACTER *character;
public:
	PLAYER(int client_id);
	~PLAYER();

	// TODO: clean this up
	char skin_name[64];
	int use_custom_color;
	int color_body;
	int color_feet;
	
	int respawn_tick;
	int die_tick;
	
	bool spawning;
	int client_id;
	int team;
	int score;
	bool force_balanced;
	
	//
	int vote;
	int64 last_votecall;

	//
	int64 last_chat;
	int64 last_setteam;
	int64 last_changeinfo;
	int64 last_emote;
	int64 last_kill;

	// network latency calculations	
	struct
	{
		int accum;
		int accum_min;
		int accum_max;
		int avg;
		int min;
		int max;	
	} latency;
	
	// this is used for snapping so we know how we can clip the view for the player
	vec2 view_pos;

	void init(int client_id);
	
	CHARACTER *get_character();
	
	void kill_character(int weapon);

	void try_respawn();
	void respawn();
	void set_team(int team);
	
	void tick();
	void snap(int snapping_client);

	void on_direct_input(NETOBJ_PLAYER_INPUT *new_input);
	void on_predicted_input(NETOBJ_PLAYER_INPUT *new_input);
	void on_disconnect();
	
	
	//EXPlorer :
	
	//bots :
	struct BOT_SPAWN * bot_spawn;
	int bot_level;
	inline bool is_bot() { return team == 1; };
	void make_bot(struct BOT_SPAWN * spawn);
	void destroy_bot();
	int max_health();
	int max_armor();
	const char * get_monster_name();
	bool must_remove_bot;
	float nobody_timer;
	
	//other/all :
	void load_game(vec2 spawnpos, int flag, int kills, int time, int armor, flags w, ITEMS items, bool b_hitter, bool b_killer);
	struct GAME
	{
		int enter_tick; //ticks
		int time; //seconds
		int kills;
		int last_flag;
		int armor_max;
		flags weapons;
		float save_timer; //ticks
		float regen_timer; //ticks
		float poison_timer; //ticks
		ITEMS items;
		bool boss_hitter;
		bool boss_killer;
	} game_exp;
	void get_weapon(int wid);
};

#endif
